"""
Nyx, blazing fast astrodynamics
Copyright (C) 2023 Christopher Rabotin <christopher.rabotin@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from dataclasses import dataclass
from pathlib import Path

import yaml
from nyx_space.time import Epoch, Unit

_default_range_noise_km = {
    "tau": "24 h",
    "bias_sigma": 5.0e-3,  # 5 m
    "steady_state_sigma": 0.1e-3,  # 0.1 m
}

_default_doppler_noise_km_s = {
    "tau": "24 h",
    "bias_sigma": 50.0e-6,  # 5 cm/s
    "steady_state_sigma": 1.5e-6,  # 0.15 cm/s
}

f_2ghz = 2.036250e9  # Hz -- the  2 GHz frequency of DORIS
f_400mhz = 250e6  # Hz -- the 400 MHz frequency of DORIS


@dataclass
class RnxMsr:
    """
    Stores a RINEX measurement
    """

    epoch: Epoch  # Includes the clock offset information
    station_name: str
    l1_phase_cycles: float  #  2GHz phase cycle data
    l2_phase_cycles: float  # 400MHz phase cycle data
    c1_pseudo_range_km: float  # 2GHz pseudo range data
    c2_pseudo_range_km: float  # 400MHz pseudo range data
    rel_freq_offset: float  # in 1e-11


def _to_dec(in_str):
    deg, min, sec = [float(x) for x in in_str.split()]
    return deg + min / 60.0 + sec / 3600.0


def snx_to_groundstation(
    snx_path,
    output_path,
    frame="IAU Earth",
    elevation_mask_deg=0.0,
    range_noise_km=_default_range_noise_km,
    doppler_noise_km_s=_default_doppler_noise_km_s,
):
    """
    Convert a SINEX file to a ground station YAML file.

    This only converts the data between the `+SITE/ID` and `-SITE/ID` lines.
    """

    offsets = {
        "name": (0, 5),
        "PT": (6, 2),
        "__DOMES__": (9, 9),
        "T": (10, 1),
        "description": (21, 22),
        "longitude_deg": (44, 11),
        "latitude_deg": (56, 11),
        "height_km": (68, 7),
    }

    entries = []

    with open(snx_path, "r") as snx_file:
        parsing = False
        for line in snx_file.readlines():
            if not parsing and line.startswith("+SITE/ID"):
                parsing = True
                continue
            elif parsing:
                if line.startswith("-SITE/ID"):
                    print(
                        "Parsed {} ground stations from {}.".format(
                            len(entries), snx_path
                        )
                    )
                    break
                elif line.startswith("*CODE"):
                    continue
                else:
                    # Parse the data line
                    kwrds = [
                        line[idx : idx + klen].strip() for idx, klen in offsets.values()
                    ]
                    # Create the YAML
                    this_entry = dict(zip(offsets.keys(), kwrds))
                    # Fix the latitude and longitude to be decimals
                    this_entry["longitude_deg"] = _to_dec(this_entry["longitude_deg"])
                    this_entry["latitude_deg"] = _to_dec(this_entry["latitude_deg"])
                    # Convert the height to km
                    this_entry["height_km"] = float(this_entry["height_km"]) / 1000.0

                    # Append the Nyx specific data
                    this_entry["frame"] = frame
                    this_entry["elevation_mask_deg"] = elevation_mask_deg
                    this_entry["range_noise_km"] = range_noise_km
                    this_entry["doppler_noise_km_s"] = doppler_noise_km_s
                    this_entry["light_time_correction"] = True
                    # Remove the unused keys
                    del this_entry["__DOMES__"]
                    del this_entry["T"]
                    del this_entry["PT"]

                    entries += [this_entry]

    output_path = str(Path(output_path).resolve())

    # Dump the YAML without creating anchors
    # cf. https://github.com/yaml/pyyaml/issues/103
    class NoAliasDumper(yaml.Dumper):
        def ignore_aliases(self, _data):
            return True

    yaml.dump(entries, open(output_path, "w"), Dumper=NoAliasDumper)
    print("Stored {} ground stations in {}.".format(len(entries), output_path))


def basic_rinex(rnx_path, stations):
    """
    Parse a RINEX file to extract the basic information.
    """

    action = "hdr"

    rnx_stations = {}
    data = []

    with open(rnx_path) as rnx_file:
        for lno, line in enumerate(rnx_file.readlines()):
            if action == "hdr":
                if line[0] == "D" and len(line.split()[0]) == 3:
                    # We've found the first station
                    action = "stations"
            if action == "stations":
                # Parse the station data
                if line.startswith(" "):
                    # End of the stations
                    action = "new_record"
                    continue
                splt = line.split()
                # The following will raise an error if the station code is not in the stations dictionary
                # Note that we ignore the dome number and the frequency shift factor
                rnx_stations[splt[0]] = stations[splt[1]]

            elif action == "msr_start":
                if line[0] == ">":
                    action = "new_record"
                else:
                    # This is a new entry in the measurement block
                    splt = line.split()
                    # Note that we set the relative frequency offset to 0.0 but we'll modify it at the next line
                    try:
                        msr = RnxMsr(
                            station_name=rnx_stations[splt[0]].name,
                            epoch=epoch,
                            l1_phase_cycles=float(splt[1]),
                            l2_phase_cycles=float(splt[2]),
                            # Documentation says the data is in units of 0.01 km
                            c1_pseudo_range_km=float(splt[3]) / 100.0,
                            c2_pseudo_range_km=float(splt[4]) / 100.0,
                            rel_freq_offset=0.0,
                        )
                    except ValueError:
                        print(f"Error at line {lno} - stopping here")
                        break
                    data += [msr]
                    action = "msr_continued"
            elif action == "msr_continued":
                splt = line.split()
                if len(splt) != 9:
                    # Something is wrong, let's remove this measurement
                    data.pop()
                    action = "msr_start"
                    continue
                rel_freq_offset = float(splt[2])
                msr = data.pop()
                msr.rel_freq_offset = rel_freq_offset
                data += [msr]
                action = "msr_start"

            if action == "new_record" and line[0] == ">":
                # Compute the epoch of this record
                splt = line.split()
                if int(splt[-1]) != 0:
                    print(f"Skipping unusable measurement block: {line}")
                    continue
                offset = Unit.Second * float(splt[-2])
                esplt = splt[1:-4]
                epoch = (
                    Epoch(
                        f"{esplt[0]}-{esplt[1]}-{esplt[2]} {esplt[3]}:{esplt[4]}:{esplt[5]} TAI"
                    )
                    + offset
                )
                action = "msr_start"

    print("Parsed {} measurements from {}.".format(len(data), rnx_path))
    return data


def prototype(gs_yaml, rnx_path, plot=True):
    import numpy as np
    import plotly.graph_objects as go
    from nyx_space.orbit_determination import GroundStation

    stations = GroundStation.load_many(gs_yaml)
    ground_stations = {}
    for station in stations:
        ground_stations[station.name] = station
    data = basic_rinex(rnx_path, ground_stations)
    gs_to_msrs = {}
    for msr in data:
        if msr.station_name not in gs_to_msrs:
            gs_to_msrs[msr.station_name] = []
        gs_to_msrs[msr.station_name] += [msr]

    # Build the time arrays one ground station at a time
    for gs_name, msrs in gs_to_msrs.items():
        time_offsets = [
            0.0,
        ]
        c1_data = []
        ref_epoch = msrs[0].epoch
        for msr in msrs[1:]:
            if msr.c1_pseudo_range_km > 0.0:
                # This data is invalid
                continue
            time_offsets += [msr.epoch.timedelta(ref_epoch).to_seconds()]
            c1_data += [msr.c1_pseudo_range_km]

        # Some preprocessing to try to make this zero meaned
        c1_data = (np.array(c1_data) - np.mean(c1_data)) / np.std(c1_data)

        try:
            c1_corr = np.correlate(time_offsets, c1_data, mode="same")
        except ValueError:
            print(f"Skipping {gs_name} due to a ValueError")
            continue

        tau_val = np.max(c1_corr) / np.exp(1)
        tau = time_offsets[np.argmin(np.abs(c1_corr - tau_val))]
        # Compute the stddev of the data after 2*tau
        c1_stddev = np.std([x for t, x in zip(time_offsets, c1_data) if t > 2 * tau])

        dt = msrs[-1].epoch.timedelta(ref_epoch)
        title = f"For {ground_stations[gs_name]}: τ = {tau} s, σ = {c1_stddev * np.std(c1_data)} km (with {len(msrs)} measurements over {dt})"
        print(title)
        print(f"\tC1: {c1_corr}")
        if len(msrs) > 100 and plot:
            fig = go.Figure()
            fig.add_trace(
                go.Scatter(
                    x=time_offsets, y=c1_corr, mode="lines", name="C1 data (2 GHz)"
                )
            )
            # Add τ
            fig.add_vline(
                x=tau,
                line_width=2,
                line_dash="dash",
                line_color="red",
                row=1,
                col=1,
            )
            fig.update_layout(title=title)
            fig.show()
            # Plot the data itself too
            fig = go.Figure()
            fig.add_trace(
                go.Scatter(
                    x=time_offsets, y=c1_data, mode="lines", name="C1 data (2 GHz)"
                )
            )
            fig.update_layout(title=title)
            fig.show()

    """
    Iterate through different true anomaly values and check which lead to the least errors
    """
    import logging

    import numpy as np
    import pandas as pd
    from nyx_space.cosmic import Cosm, Orbit, Spacecraft
    from nyx_space.mission_design import (
        TrajectoryLoader,
        SpacecraftDynamics,
        propagate,
    )
    from nyx_space.orbit_determination import GroundTrackingArcSim, TrkConfig

    FORMAT = "%(levelname)s %(name)s %(asctime)-15s %(filename)s:%(lineno)d %(message)s"
    logging.basicConfig(format=FORMAT)
    logging.getLogger().setLevel(logging.INFO)

    cosm = Cosm.de438()
    eme2k = cosm.frame("EME2000")

    # Base path
    root = Path(__file__).joinpath("../../../../").resolve()
    config_path = root.joinpath("./data/tests/config/")
    outpath = root.joinpath("output_data/")

    # Load the dynamics
    dynamics = SpacecraftDynamics.load_named(str(config_path.joinpath("dynamics.yaml")))

    for gs_name, msrs in gs_to_msrs.items():
        if len(msrs) < 100:
            continue
        # Build the approximate orbit
        # We'll use the first measurement as the reference epoch
        ref_epoch = msrs[0].epoch

        # We'll use the first measurement as the reference station
        ref_station = stations[gs_name]
        # Setup a continuous tracking for this station
        trk_cfg = {ref_station.name: TrkConfig()}

        for ta_deg in np.linspace(0, 360, 10):
            orbit = Orbit.from_keplerian(
                sma_km=7349136.3e-3,
                ecc=0.00117,
                inc_deg=99.35,
                raan_deg=108.48,
                aop_deg=110.9766,
                ta_deg=ta_deg,
                epoch=ref_epoch,
                frame=eme2k,
            )

            # Build a spacecraft
            sc = Spacecraft(
                orbit,
                dry_mass_kg=1500.0,
                fuel_mass_kg=0.0,
            )

            # Propagate this trajectory for one day
            # An propagate for two periods (we only care about the trajectory)
            _, traj = propagate(sc, dynamics["hifi"], Unit.Day * 1)
            print(traj)

            # This is a bit of a pain but at the moment, we must serialize the trajectory to a file
            # and reload it into a Dynamic trajectory before passing it to the Python object.

            traj_file = str(outpath.joinpath("./od_val_with_arc_truth_ephem.parquet"))
            traj.to_parquet(traj_file)
            traj = TrajectoryLoader(traj_file)

            # Build a tracking arc
            arc_sim = GroundTrackingArcSim([ref_station], traj, trk_cfg, 0)
            # Generate the measurements
            path = arc_sim.generate_measurements(
                str(outpath.joinpath(f"./from_{ref_station.name}-ta_{ta_deg}.parquet")),
                timestamp=False,
                metadata={},
            )
            # Load this parquet and print the head
            df = pd.read_parquet(path)
            print(df.head())

            break
        break



"""
Nyx, blazing fast astrodynamics
Copyright (C) 2023 Christopher Rabotin <christopher.rabotin@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import plotly.graph_objects as go
from datetime import datetime
import numpy as np

# Color blindness friendly colors, from https://jfly.uni-koeln.de/color/#pallet
# And from https://davidmathlogic.com/colorblind/#%23E1BE6A-%2340B0A6-%237bcb12-%233d9087
colors = {
    "orange": (230, 159, 0),  # #e69f00
    "sky_blue": (86, 180, 233),  # #56b4e9
    "green": (0, 158, 115),  # #009e73
    "yellow": (240, 228, 66),  # #f0e442
    "blue": (0, 114, 178),  # #0072b2
    "red_ish": (213, 94, 0),  # #d55e00
    "purple": (204, 121, 167),  # #cc79a7
    "yellow2": (255, 194, 10),  # #ffc20a
    "blue2": (12, 123, 220),  # #0c7bdc
    "bright_green": (26, 255, 26),  # #1aff1a
    "purple2": (75, 0, 146),  # #4b0092
    "brown": (153, 79, 0),  # #994f00
    "blue3": (0, 108, 209),  # #006cd1
    "light_yellow": (254, 254, 98),  # #fefe62
    "pink": (211, 95, 183),  # #d35fb7
    "black": (0, 0, 0),
    "gray": (231, 231, 231),  # #e7e7e7
    "extra-1": (216, 27, 96),  # #D81B60
    "extra-2": (30, 136, 229),  # #1E88E5
    "extra-4": (0, 77, 64),  # #004D40
    "extra-5": (215, 106, 161),  # #d76aa1
    "extra-6": (46, 18, 240),  # #2e12f0
    "extra-7": (190, 222, 214),  # #beded6
    "extra-8": (27, 179, 109),  # #1bb36d
    "extra-9": (232, 17, 141),  # #e8118d
    "extra-10": (27, 33, 247),  # #1b21f7
    "extra-3": (255, 193, 7),  # #FFC107
}

# Radii of different celestial bodies in km
radii = {
    "Sun": 696342.000000,
    "Mercury": 2439.700000,
    "Venus": 6051.900000,
    "Earth": 6378.136300,
    "Luna": 1738.100000,
    "Mars": 3397.000000,
    "Jupiter": 71492.000000,
    "Saturn": 60268.000000,
    "Uranus": 25559.000000,
    "Neptune": 25269.000000,
}

body_color = {
    "Sun": colors["yellow"],
    "Mercury": colors["extra-7"],
    "Venus": colors["orange"],
    "Earth": colors["sky_blue"],
    "Luna": colors["gray"],
    "Mars": colors["red_ish"],
    "Jupiter": colors["yellow2"],
    "Saturn": colors["extra-3"],
    "Uranus": colors["extra-2"],
    "Neptune": colors["extra-10"],
}


def _add_watermark(who):
    nyx_tpl = go.layout.Template()
    year = datetime.now().year
    nyx_tpl.layout.annotations = [
        dict(
            name="watermark",
            text=f"{who} © {year}",
            textangle=0,
            opacity=0.1,
            font=dict(color="black", size=100),
            xref="paper",
            yref="paper",
            x=0.1,
            y=0.1,
            showarrow=False,
        )
    ]


year = datetime.now().year
nyx_tpl = go.layout.Template()
nyx_tpl.layout.annotations = [
    dict(
        name="watermark",
        text=f"Nyx Space 🄯 AGPLv3 {year}",
        opacity=0.75,
        font=dict(color="#3d84e8", size=12),
        xref="paper",
        yref="paper",
        x=0.95,
        y=0.08,
        showarrow=False,
    )
]


def plot_with_error(
    fig, df, x, y_name, err_name, color_name, xtitle, ytitle, title, copyright=None
):
    """
    Adds the traces for the provided X vs Y with ERR around Y.
    """
    color = colors[color_name]

    fig.add_trace(
        go.Scatter(
            name=y_name,
            x=x,
            y=df[y_name],
            mode="lines",
            line=dict(color=f"rgb({color[0]}, {color[1]}, {color[2]})"),
        )
    )

    fig.add_trace(
        go.Scatter(
            name=f"+{y_name}",
            x=x,
            y=df[y_name] + df[err_name],
            mode="lines",
            marker=dict(
                color=f"rgb({int(color[0]*0.85)}, {int(color[1]*0.85)}, {int(color[2]*0.85)})"
            ),
            line=dict(width=0),
            showlegend=False,
        )
    )
    fig.add_trace(
        go.Scatter(
            name=f"-{y_name}",
            x=x,
            y=df[y_name] - df[err_name],
            marker=dict(
                color=f"rgb({int(color[0]*0.85)}, {int(color[1]*0.85)}, {int(color[2]*0.85)})"
            ),
            line=dict(width=0),
            mode="lines",
            fillcolor=f"rgba({color[0]}, {color[1]}, {color[1]}, 0.3)",
            fill="tonexty",
            showlegend=False,
        )
    )

    finalize_plot(fig, title, xtitle, ytitle, copyright)


def plot_line(fig, df, x, y_name, color_name, xtitle, ytitle, title, copyright):
    """
    Adds the trace for the provided X vs Y.
    """

    plot_raw_line(
        fig, x, df[y_name], y_name, color_name, xtitle, ytitle, title, copyright
    )


def plot_raw_line(fig, x, y, y_name, color_name, xtitle, ytitle, title, copyright=None):
    """
    Adds the trace for the provided X vs Y.
    """
    color = colors[color_name]

    fig.add_trace(
        go.Scatter(
            name=y_name,
            x=x,
            y=y,
            mode="lines",
            line=dict(color=f"rgb({color[0]}, {color[1]}, {color[2]})"),
        )
    )

    finalize_plot(fig, title, xtitle, ytitle, copyright)


def finalize_plot(fig, title, xtitle=None, ytitle=None, copyright=None):
    """
    Add titles, copyright and watermark
    """

    annotations = [dict(templateitemname="watermark")]
    if copyright:
        annotations += [
            dict(
                templateitemname="watermark",
                text=f"{copyright} © {year}",
                font=dict(color="black", size=16),
                y=0.1,
            )
        ]

    if xtitle:
        fig.update_layout(xaxis_title=xtitle)
    if ytitle:
        fig.update_layout(yaxis_title=ytitle)

    fig.update_layout(
        title=title,
        hovermode="x",
        template=nyx_tpl,
        annotations=annotations,
    )


def build_sphere(size, color, num_points=500, opacity=1.0):
    """
    Source: https://python.plainenglish.io/how-to-create-a-3d-model-of-the-solar-system-with-plotly-in-python-2a74e672b771
    """

    theta = np.linspace(0, 2 * np.pi, num_points)
    phi = np.linspace(0, np.pi, num_points)

    # Set up coordinates for points on the sphere
    x0 = size * np.outer(np.cos(theta), np.sin(phi))
    y0 = size * np.outer(np.sin(theta), np.sin(phi))
    z0 = size * np.outer(np.ones(num_points), np.cos(phi))

    # Set up trace
    trace = go.Surface(
        x=x0,
        y=y0,
        z=z0,
        colorscale=[
            [0, f"rgba({int(color[0])}, {int(color[1])}, {int(color[2])}, {opacity})"],
            [1, f"rgba({int(color[0])}, {int(color[1])}, {int(color[2])}, {opacity})"],
        ],
        hoverinfo="none",
    )
    trace.update(showscale=False)

    return trace


def build_3dline(x, y, z, text, color=colors["sky_blue"], width=2, name=None):
    """
    Super simple, but allows consistency
    """

    trace = go.Scatter3d(
        x=x,
        y=y,
        z=z,
        marker=dict(size=0.1),
        line=dict(
            color=f"rgb({int(color[0])}, {int(color[1])}, {int(color[2])})",
            width=width,
        ),
        text=text,
        name=name,
    )
    return trace



"""
Nyx, blazing fast astrodynamics
Copyright (C) 2023 Christopher Rabotin <christopher.rabotin@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import plotly.graph_objects as go

from .utils import plot_with_error, plot_line, finalize_plot, colors

import re
from slugify import slugify

import pandas as pd

import numpy as np
from scipy.stats import norm


def plot_estimates(
    dfs,
    title,
    cov_frame="RIC",
    cov_fmt="sqrt",
    cov_sigma=3.0,
    msr_df=None,
    time_col_name="Epoch:Gregorian UTC",
    ref_traj=None,
    html_out=None,
    copyright=None,
    pos_fig=None,
    vel_fig=None,
    show=True,
):
    """
    Plots the estimates from an orbit determination solution

    Args:
        dfs (pandas.DataFrame): The data frame containing the orbit determination solution (or a list thereof)
        title (str): The title of the plot
        cov_frame (str): Frame in which to plot the covariance, defaults to RIC frame. File also contains the integration frame.
        cov_fmt (str): Defines how to plot the covariance, defaults to "sqrt" (applies np.sqrt to the column), which will convert the covariance from an expectation to an uncertainty.
        cov_sigma (float): Defines the number fo sigmas of uncertainty or covariance to plot, uses [68, 95, 99.7% rule](https://en.wikipedia.org/wiki/68%E2%80%9395%E2%80%9399.7_rule) for 1.0, 2.0, and 3.0 of sigma.
        msr_df (pandas.DataFrame): The data frame containing the measurements
        time_col_name (str): The name of the time column
        ref_trajs (pandas.DataFrame): If provided, the difference between the reference trajectory and the estimates will be plotted instead of the estimates alone (or list thereof).
        html_out (str): The name of the HTML file to save the plot to
        copyright (str): The copyright to display on the plot
        pos_fig (plotly.graph_objects.Figure): The figure to plot the position estimates on
        vel_fig (plotly.graph_objects.Figure): The figure to plot the velocity estimates on
        show (bool): Whether to show the plot. If set to false, the figure will be returned.
    """

    if not isinstance(dfs, list):
        dfs = [dfs]

    if ref_traj is not None and not isinstance(ref_traj, list):
        ref_traj = [ref_traj]

    if pos_fig is None:
        pos_fig = go.Figure()

    if vel_fig is None:
        vel_fig = go.Figure()

    for num, df in enumerate(dfs):
        try:
            orig_tim_col = df[time_col_name]
        except KeyError:
            # Find the time column
            try:
                col_name = [x for x in df.columns if x.startswith("Epoch")][0]
            except IndexError:
                raise KeyError("Could not find any Epoch column")
            print(f"Could not find time column {time_col_name}, using `{col_name}`")
            orig_tim_col = df[col_name]

        # Build a Python datetime column
        pd_ok_epochs = []
        for epoch in orig_tim_col:
            epoch = epoch.replace("UTC", "").strip()
            if "." not in epoch:
                epoch += ".0"
            pd_ok_epochs += [epoch]
        time_col = pd.to_datetime(pd_ok_epochs)
        x_title = "Epoch {}".format(time_col_name[-3:])

        # Check that the requested covariance frame exists
        frames = set(
            [
                re.search(r"\((.*?)\)", c).group(1)
                for c in df.columns
                if c.startswith("Covariance")
            ]
        )
        if cov_frame not in frames:
            raise ValueError(
                f"Covariance frame `{cov_frame}` not in one of the available frames from dataframe: {frames}"
            )
        covar = {}

        # Reference the covariance frames
        for covar_var, covar_col in {
            "cx_x": "Covariance XX",
            "cy_y": "Covariance YY",
            "cz_z": "Covariance ZZ",
            "cx_dot_x_dot": "Covariance VxVx",
            "cy_dot_y_dot": "Covariance VyVy",
            "cz_dot_z_dot": "Covariance VzVz",
        }.items():
            # Create a new column with the transformed covariance. (e.g. "Covariance VzVz (RIC) 1.0-sigma sqrt")
            if cov_fmt is None:
                cov_col_name = f"{covar_col} ({cov_frame}) {cov_sigma}-sigma"
                # No transformation here
                df[cov_col_name] = df[f"{covar_col} ({cov_frame})"]
            else:
                cov_col_name = f"{covar_col} ({cov_frame}) {cov_sigma}-sigma {cov_fmt}"
                # Transform the current column
                df[cov_col_name] = eval(f"np.{cov_fmt}")(
                    df[f"{covar_col} ({cov_frame})"]
                )
            covar[f"{covar_var}"] = cov_col_name

        plt_df = df
        plt_columns = [
            "x (km)",
            "y (km)",
            "z (km)",
            "vx (km/s)",
            "vy (km/s)",
            "vz (km/s)",
        ]
        plt_modifier = ""

        if ref_traj is not None:
            # Merge both dataframes, keeping all of the values
            merged_df = pd.merge(
                df,
                ref_traj[num],
                on=time_col_name,
                how="outer",
                suffixes=("_od", "_ref_traj"),
            )
            if merged_df.shape[0] > max(df.shape[0], ref_traj[num].shape[0]):
                print(
                    "Reference epochs differ between dataframes, performing a merge by closest key on TAI epoch instead"
                )
                merged_df = pd.merge_asof(
                    df, ref_traj[num], on="Epoch:TAI (s)", suffixes=("_od", "_ref_traj")
                )

            # Add the difference to reference to the dataframe
            for coord in plt_columns:
                merged_df[f"delta {coord}"] = (
                    merged_df[f"{coord}_od"] - merged_df[f"{coord}_ref_traj"]
                )
            # Set the plotted dataframe to this one
            plt_df = merged_df
            plt_columns = [f"delta {c}" for c in plt_columns]
            plt_modifier = " error"

        plot_with_error(
            pos_fig,
            plt_df,
            time_col,
            plt_columns[0],
            covar["cx_x"],
            "blue",
            x_title,
            f"Position estimate {plt_modifier}(km)",
            f"Position estimates: {title}",
            copyright,
        )
        plot_with_error(
            pos_fig,
            plt_df,
            time_col,
            plt_columns[1],
            covar["cy_y"],
            "green",
            x_title,
            f"Position estimate {plt_modifier}(km)",
            f"Position estimates: {title}",
            copyright,
        )
        plot_with_error(
            pos_fig,
            plt_df,
            time_col,
            plt_columns[2],
            covar["cz_z"],
            "orange",
            x_title,
            f"Position estimate {plt_modifier}(km)",
            f"Position estimates: {title}",
            copyright,
        )

        plot_with_error(
            vel_fig,
            plt_df,
            time_col,
            plt_columns[3],
            covar["cx_dot_x_dot"],
            "blue",
            x_title,
            f"Velocity estimate {plt_modifier}(km/s)",
            f"Velocity estimates: {title}",
            copyright,
        )
        plot_with_error(
            vel_fig,
            plt_df,
            time_col,
            plt_columns[4],
            covar["cy_dot_y_dot"],
            "green",
            x_title,
            f"Velocity estimate {plt_modifier}(km/s)",
            f"Velocity estimates: {title}",
            copyright,
        )
        plot_with_error(
            vel_fig,
            plt_df,
            time_col,
            plt_columns[5],
            covar["cz_dot_z_dot"],
            "orange",
            x_title,
            f"Velocity estimate {plt_modifier}(km/s)",
            f"Velocity estimates: {title}",
            copyright,
        )

    if msr_df is not None:
        # Plot the measurements on both plots
        pos_fig = plot_measurements(
            msr_df, title, time_col_name, fig=pos_fig, show=False
        )

        vel_fig = plot_measurements(
            msr_df, title, time_col_name, fig=vel_fig, show=False
        )

    if html_out:
        html_out = html_out.replace(".html", "_{}.html")
        this_output = html_out.format("pos_est")
        with open(this_output, "w") as f:
            f.write(pos_fig.to_html())
        print(f"Saved HTML to {this_output}")

        this_output = html_out.format("vel_est")
        with open(this_output, "w") as f:
            f.write(vel_fig.to_html())
        print(f"Saved HTML to {this_output}")

    if show:
        pos_fig.show()
        vel_fig.show()
    else:
        return pos_fig, vel_fig


def plot_covar(
    dfs,
    title,
    cov_frame="RIC",
    cov_fmt="sqrt",
    cov_sigma=3.0,
    msr_df=None,
    time_col_name="Epoch:Gregorian UTC",
    html_out=None,
    copyright=None,
    pos_fig=None,
    vel_fig=None,
    show=True,
):
    """
    Plot only the covariance from an orbit determination solution

    Args:
        dfs (pandas.DataFrame): The data frame containing the orbit determination solution (or a list thereof)
        title (str): The title of the plot
        cov_frame (str): Frame in which to plot the covariance, defaults to RIC frame. File also contains the integration frame.
        cov_fmt (str): Defines how to plot the covariance, defaults to "sqrt" (applies np.sqrt to the column), which will convert the covariance from an expectation to an uncertainty.
        cov_sigma (float): Defines the number fo sigmas of uncertainty or covariance to plot, uses [68, 95, 99.7% rule](https://en.wikipedia.org/wiki/68%E2%80%9395%E2%80%9399.7_rule) for 1.0, 2.0, and 3.0 of sigma.
        msr_df (pandas.DataFrame): The data frame containing the measurements
        time_col_name (str): The name of the time column
        html_out (str): The name of the HTML file to save the plot to
        copyright (str): The copyright to display on the plot
        pos_fig (plotly.graph_objects.Figure): The figure to plot the position estimates on
        vel_fig (plotly.graph_objects.Figure): The figure to plot the velocity estimates on
        show (bool): Whether to show the plot. If set to false, the figure will be returned.
    """

    if not isinstance(dfs, list):
        dfs = [dfs]

    if pos_fig is None:
        pos_fig = go.Figure()

    if vel_fig is None:
        vel_fig = go.Figure()

    for df in dfs:
        try:
            orig_tim_col = df[time_col_name]
        except KeyError:
            # Find the time column
            try:
                col_name = [x for x in df.columns if x.startswith("Epoch")][0]
            except IndexError:
                raise KeyError("Could not find any Epoch column")
            print(f"Could not find time column {time_col_name}, using `{col_name}`")
            orig_tim_col = df[col_name]

        # Build a Python datetime column
        pd_ok_epochs = []
        for epoch in orig_tim_col:
            epoch = epoch.replace("UTC", "").strip()
            if "." not in epoch:
                epoch += ".0"
            pd_ok_epochs += [epoch]
        time_col = pd.to_datetime(pd_ok_epochs)
        x_title = "Epoch {}".format(time_col_name[-3:])

        # Check that the requested covariance frame exists
        frames = set(
            [
                re.search(r"\((.*?)\)", c).group(1)
                for c in df.columns
                if c.startswith("Covariance")
            ]
        )
        if cov_frame not in frames:
            raise ValueError(
                f"Covariance frame `{cov_frame}` not in one of the available frames from dataframe: {frames}"
            )
        covar = {}

        # Reference the covariance frames
        for covar_var, covar_col in {
            "cx_x": "Covariance XX",
            "cy_y": "Covariance YY",
            "cz_z": "Covariance ZZ",
            "cx_dot_x_dot": "Covariance VxVx",
            "cy_dot_y_dot": "Covariance VyVy",
            "cz_dot_z_dot": "Covariance VzVz",
        }.items():
            # Create a new column with the transformed covariance. (e.g. "Covariance VzVz (RIC) 1.0-sigma sqrt")
            cov_col_name = f"{covar_col} ({cov_frame}) {cov_sigma}-sigma {cov_fmt}"
            # Transform the current column
            df[cov_col_name] = eval(f"np.{cov_fmt}")(df[f"{covar_col} ({cov_frame})"])
            covar[f"{covar_var}"] = cov_col_name

        plot_line(
            pos_fig,
            df,
            time_col,
            covar["cx_x"],
            "blue",
            x_title,
            f"Position covariance {cov_sigma}-sigma (km)",
            title,
            copyright,
        )
        plot_line(
            pos_fig,
            df,
            time_col,
            covar["cy_y"],
            "green",
            x_title,
            f"Position covariance {cov_sigma}-sigma (km)",
            title,
            copyright,
        )
        plot_line(
            pos_fig,
            df,
            time_col,
            covar["cz_z"],
            "orange",
            x_title,
            f"Position covariance {cov_sigma}-sigma (km)",
            title,
            copyright,
        )

        # Autoscale
        hwpt = int(len(df) / 2)
        max_cov_y = max(
            max(df[covar["cx_x"]][hwpt:]),
            max(df[covar["cy_y"]][hwpt:]),
            max(df[covar["cz_z"]][hwpt:]),
        )
        pos_fig.update_layout(yaxis_range=[-0.1 * max_cov_y, 1.5 * max_cov_y])

        plot_line(
            vel_fig,
            df,
            time_col,
            covar["cx_dot_x_dot"],
            "blue",
            x_title,
            f"Velocity covariance {cov_sigma}-sigma (km/s)",
            title,
            copyright,
        )
        plot_line(
            vel_fig,
            df,
            time_col,
            covar["cy_dot_y_dot"],
            "green",
            x_title,
            f"Velocity covariance {cov_sigma}-sigma (km/s)",
            title,
            copyright,
        )
        plot_line(
            vel_fig,
            df,
            time_col,
            covar["cz_dot_z_dot"],
            "orange",
            x_title,
            f"Velocity covariance {cov_sigma}-sigma (km/s)",
            title,
            copyright,
        )
        # Autoscale
        hwpt = int(len(df) / 2)
        max_cov_y = max(
            max(df[covar["cx_dot_x_dot"]][hwpt:]),
            max(df[covar["cy_dot_y_dot"]][hwpt:]),
            max(df[covar["cz_dot_z_dot"]][hwpt:]),
        )

        vel_fig.update_layout(yaxis_range=[-0.1 * max_cov_y, 1.5 * max_cov_y])

    if msr_df is not None:
        # Plot the measurements on both plots
        pos_fig = plot_measurements(
            msr_df, title, time_col_name, fig=pos_fig, show=False
        )

        vel_fig = plot_measurements(
            msr_df, title, time_col_name, fig=vel_fig, show=False
        )

    if html_out:
        html_out = html_out.replace(".html", "_{}.html")
        this_output = html_out.format("pos_cov")
        with open(this_output, "w") as f:
            f.write(pos_fig.to_html())
        print(f"Saved HTML to {this_output}")

        this_output = html_out.format("vel_cov")
        with open(this_output, "w") as f:
            f.write(vel_fig.to_html())
        print(f"Saved HTML to {this_output}")

    if show:
        pos_fig.show()
        vel_fig.show()
    else:
        return pos_fig, vel_fig


def plot_measurements(
    dfs,
    title,
    time_col_name="Epoch:Gregorian UTC",
    html_out=None,
    copyright=None,
    fig=None,
    show=True,
):
    if not isinstance(dfs, list):
        dfs = [dfs]

    if fig is None:
        fig = go.Figure()

    color_values = list(colors.values())

    station_colors = {}

    for df in dfs:
        try:
            orig_tim_col = df[time_col_name]
        except KeyError:
            # Find the time column
            try:
                col_name = [x for x in df.columns if x.startswith("Epoch")][0]
            except IndexError:
                raise KeyError("Could not find any Epoch column")
            print(f"Could not find time column {time_col_name}, using `{col_name}`")
            orig_tim_col = df[col_name]

        # Build a Python datetime column
        pd_ok_epochs = []
        for epoch in orig_tim_col:
            epoch = epoch.replace("UTC", "").strip()
            if "." not in epoch:
                epoch += ".0"
            pd_ok_epochs += [epoch]
        time_col = pd.to_datetime(pd_ok_epochs)
        x_title = "Epoch {}".format(time_col_name[-3:])

        # Diff the epochs of the measurements to find when there is a start and end.

        # Get Epoch TAI diffs of diffs
        diff2 = df["Epoch:TAI (s)"].diff().diff()

        # Lists to hold segments
        station_names = []
        start_indices = []
        end_indices = []

        # Loop through diff2 values
        for i, val in enumerate(diff2):
            # Positive spike indicates end of station segment at the prior index and the start of a new segment
            if val > 0:
                end_idx = i - 1
                end_indices.append(end_idx)
                start_idx = i
                start_indices.append(start_idx)
                station_names.append(df.loc[start_idx, "Tracking device"])
        # Add the start and end data
        start_indices = [0] + start_indices
        end_indices += [len(df) - 1]
        station_names = [df["Tracking device"][0]] + station_names

        # Zip into segments
        station_segments = list(zip(station_names, start_indices, end_indices))

        # Add tracking passes to the plot
        for ii, (name, start_idx, end_idx) in enumerate(station_segments):
            start_epoch = time_col[start_idx]
            end_epoch = time_col[end_idx]

            try:
                color = station_colors[name]
            except KeyError:
                color = color_values[(len(fig.data) + ii) % len(color_values)]
                color = f"rgb({int(color[0])}, {int(color[1])}, {int(color[2])})"
                station_colors[name] = color

            fig.add_vrect(
                x0=start_epoch,
                x1=end_epoch,
                annotation_text=name,
                annotation_position="top left",
                fillcolor=color,
                opacity=0.25,
                line_width=0,
            )

        finalize_plot(fig, title, x_title, copyright, show)

    if html_out:
        with open(html_out, "w") as f:
            f.write(fig.to_html())
        print(f"Saved HTML to {html_out}")

    if show:
        fig.show()
    else:
        return fig


def plot_residuals(
    df,
    title,
    kind="Prefit",
    time_col_name="Epoch:Gregorian UTC",
    msr_df=None,
    copyright=None,
    html_out=None,
    show=True,
):
    """
    Plot of residuals, with 3-σ lines
    """

    try:
        orig_tim_col = df[time_col_name]
    except KeyError:
        # Find the time column
        try:
            col_name = [x for x in df.columns if x.startswith("Epoch")][0]
        except IndexError:
            raise KeyError("Could not find any Epoch column")
        print(f"Could not find time column {time_col_name}, using `{col_name}`")
        orig_tim_col = df[col_name]

    # Build a Python datetime column
    pd_ok_epochs = []
    for epoch in orig_tim_col:
        epoch = epoch.replace("UTC", "").strip()
        if "." not in epoch:
            epoch += ".0"
        pd_ok_epochs += [epoch]
    time_col = pd.to_datetime(pd_ok_epochs)
    x_title = "Epoch {}".format(time_col_name[-3:])

    plt_any = False

    for col in df.columns:
        if col.startswith(kind):
            fig = go.Figure()
            residuals = df[col]

            # Add scatter trace of residual vs normal quantiles
            fig.add_trace(
                go.Scatter(
                    x=time_col,
                    y=residuals,
                    mode="markers",
                    name=f"Residuals {col}",
                )
            )

            mean = np.mean(residuals)
            std = np.std(residuals)
            # Add the 3-σ lines
            three_sig_color = colors["red_ish"]
            three_sig_color = f"rgb({int(three_sig_color[0])}, {int(three_sig_color[1])}, {int(three_sig_color[2])})"
            fig.add_hline(
                y=mean + 3 * std,
                line_dash="dash",
                line_color=three_sig_color,
                annotation_text="+ 3σ",
            )
            fig.add_hline(
                y=mean - 3 * std,
                line_dash="dash",
                line_color=three_sig_color,
                annotation_text="- 3σ",
            )
            # Add the 1-σ lines
            one_sig_color = colors["bright_green"]
            one_sig_color = f"rgb({int(one_sig_color[0])}, {int(one_sig_color[1])}, {int(one_sig_color[2])})"
            fig.add_hline(
                y=mean + std,
                line_dash="dot",
                line_color=one_sig_color,
                annotation_text="+ 1σ",
            )
            fig.add_hline(
                y=mean - std,
                line_dash="dot",
                line_color=one_sig_color,
                annotation_text="- 1σ",
            )

            if msr_df is not None:
                # Plot the measurements on both plots
                fig = plot_measurements(
                    msr_df, title, time_col_name, fig=fig, show=False
                )

            finalize_plot(
                fig, title=f"{title} {col}", xtitle=x_title, copyright=copyright
            )

            plt_any = True

            if html_out:
                this_output = html_out.replace(".html", f"_{slugify(col)}.html")
                with open(this_output, "w") as f:
                    f.write(fig.to_html())
                print(f"Saved HTML to {this_output}")

            if show:
                fig.show()

    if not plt_any:
        raise ValueError(f"No columns ending with {kind} found -- nothing plotted")


def plot_residual_histogram(
    df, title, kind="Prefit", copyright=None, html_out=None, show=True
):
    """
    Histogram of residuals
    """

    for col in df.columns:
        if col.startswith(kind):
            residuals = df[col]
            fig = go.Figure()
            fig.add_trace(
                go.Histogram(
                    x=residuals,
                    name=f"Histogram of Residuals for {col}",
                )
            )

            # Add normal distribution with same mean and std
            mean = np.mean(residuals)
            std = np.std(residuals)
            x = np.linspace(mean - 3 * std, mean + 3 * std, 100)
            pdf = norm.pdf(x, mean, std)
            fig.add_trace(
                go.Scatter(
                    x=x,
                    y=pdf,
                    name=f"Normal Distribution for {col}",
                    line_dash="dash",
                    line_color="red",
                )
            )

            finalize_plot(fig, title, xtitle=None, copyright=copyright)

            if html_out:
                this_output = html_out.replace(".html", f"_{slugify(col)}.html")
                with open(this_output, "w") as f:
                    f.write(fig.to_html())
                print(f"Saved HTML to {this_output}")

            if show:
                fig.show()



from nyx_space.cosmic import Cosm, Orbit, Spacecraft, SrpConfig, DragConfig
from nyx_space.time import Epoch, Unit, Duration
from nyx_space.monte_carlo import generate_orbits, generate_spacecraft, StateParameter


def test_load_cosm():
    cosm = Cosm.de438()
    assert cosm.frames_get_names() == [
        "SSB J2000",
        "Sun J2000",
        "iau sun",
        "Mercury Barycenter J2000",
        "Venus Barycenter J2000",
        "iau venus",
        "Earth Barycenter J2000",
        "Earth J2000",
        "iau earth",
        "Moon J2000",
        "iau moon",
        "Mars Barycenter J2000",
        "iau mars",
        "Jupiter Barycenter J2000",
        "iau jupiter",
        "Saturn Barycenter J2000",
        "iau saturn",
        "Uranus Barycenter J2000",
        "iau uranus",
        "Neptune Barycenter J2000",
        "iau neptune",
        "Pluto Barycenter J2000",
    ]


def test_define_spacecraft():
    cosm = Cosm.de438()
    eme2k = cosm.frame("EME2000")

    assert eme2k.gm() == 398600.435392  # SPICE data, not GMAT data (398600.4415)
    assert eme2k.is_geoid()
    assert eme2k.equatorial_radius() == 6378.1363

    e = Epoch.system_now()

    orbit = Orbit.from_keplerian_altitude(
        400,
        ecc=1e-4,
        inc_deg=30.5,
        raan_deg=35.0,
        aop_deg=65.0,
        ta_deg=590,
        epoch=e,
        frame=eme2k,
    )

    assert orbit.period() == Unit.Second * 5553.623455582

    print(orbit)
    print(repr(orbit))

    # Define the SRP
    srp = SrpConfig(2.0)
    # Define the Drag
    drag = DragConfig(2.0)

    # Build a spacecraft
    sc = Spacecraft(
        orbit,
        dry_mass_kg=500.0,
        fuel_mass_kg=15.0,
        srp=srp,
        drag=drag,
    )

    print(sc)

    print(sc.orbit)

    # NOTE: This does not return a pointer to the object, but a new object!
    # Therefore the equality _will_ fail!
    assert orbit != sc.orbit
    # But the printed information is identical
    assert f"{orbit}" == f"{sc.orbit}"
    # And the epochs match
    assert e.timedelta(sc.epoch) == Duration.zero()


def test_generate_states():
    """
    Tests that we can generate orbits from their state parameter deviations
    """
    # Build a demo orbit
    cosm = Cosm.de438()
    eme2k = cosm.frame("EME2000")

    assert eme2k.gm() == 398600.435392  # SPICE data, not GMAT data (398600.4415)
    assert eme2k.is_geoid()
    assert eme2k.equatorial_radius() == 6378.1363

    e = Epoch.system_now()

    orbit = Orbit.from_keplerian_altitude(
        400,
        ecc=1e-4,
        inc_deg=30.5,
        raan_deg=35.0,
        aop_deg=65.0,
        ta_deg=590,
        epoch=e,
        frame=eme2k,
    )

    # Generate a bunch of them in percentage of error
    orbits = generate_orbits(
        orbit,
        [
            # Note that we can create the state parameter from a string of its name
            (StateParameter("SMA"), 0.05),
            # Or directly as an enum (the preferred method)
            (StateParameter.Eccentricity, 0.1),
            (StateParameter.Inclination, 0.1),
        ],
        100,
        kind="prct",
    )
    assert len(orbits) == 100

    # Define the SRP
    srp = SrpConfig(2.0)
    # Define the Drag
    drag = DragConfig(2.0)

    # Build a spacecraft
    sc = Spacecraft(
        orbit,
        dry_mass_kg=500.0,
        fuel_mass_kg=15.0,
        srp=srp,
        drag=drag,
    )

    # Generate a bunch of spacecraft with a specific seed, and using the absolute errors
    spacecraft = generate_spacecraft(
        sc,
        [
            (StateParameter.FuelMass, 1.9),
            (StateParameter.SMA, 1.0),
            (StateParameter.Eccentricity, 0.01),
            (StateParameter.Inclination, 3.5),
        ],
        100,
        kind="abs",
        seed=42,
    )
    assert len(spacecraft) == 100



import logging
from pathlib import Path
import numpy as np
import pandas as pd
import sys

from nyx_space.orbit_determination import (
    GroundStation,
    GroundTrackingArcSim,
    TrkConfig,
    OrbitEstimate,
    process_tracking_arc,
    predictor,
    DynamicTrackingArc,
    ExportCfg,
)
from nyx_space.mission_design import TrajectoryLoader, SpacecraftDynamics, propagate
from nyx_space.cosmic import Spacecraft
from nyx_space.time import Unit
from nyx_space.plots.od import (
    plot_covar,
    plot_estimates,
    plot_residuals,
    plot_residual_histogram,
)
from nyx_space.plots.traj import plot_orbit_elements


def test_filter_arc():
    # Initialize logging
    FORMAT = "%(levelname)s %(name)s %(asctime)-15s %(filename)s:%(lineno)d %(message)s"
    logging.basicConfig(format=FORMAT)
    logging.getLogger().setLevel(logging.INFO)

    # Base path
    root = Path(__file__).joinpath("../../../").resolve()
    config_path = root.joinpath("./data/tests/config/")
    outpath = root.joinpath("output_data/")

    # Load the dynamics and spacecraft
    sc = Spacecraft.load(str(config_path.joinpath("spacecraft.yaml")))
    dynamics = SpacecraftDynamics.load_named(str(config_path.joinpath("dynamics.yaml")))

    # An propagate for two periods (we only care about the trajectory)
    _, traj = propagate(sc, dynamics["hifi"], sc.orbit.period() * 2)
    # Resample the trajectory at fixed step size
    traj = traj.resample(Unit.Second * 10.0)
    # And save the trajectory
    traj_file = str(outpath.joinpath("./python_ref_traj.parquet"))
    traj.to_parquet(traj_file)

    # Now starts the measurement generation

    # Load the devices
    devices = GroundStation.load_many(
        str(root.joinpath("./data/tests/config/many_ground_stations.yaml"))
    )
    print(f"Loaded {devices}")
    for device in devices:
        # Check that we have loaded them correctly
        if device.name == "Demo ground station":
            assert device.latitude_deg == 2.3522
            assert device.longitude_deg == 48.8566
        else:
            # This is Canberra, inherited from the build-in ground station
            assert device.latitude_deg == -35.398333
            assert device.longitude_deg == 148.981944
        assert device.elevation_mask_deg == 5.0  # Both have the same mask

    # Load the tracking configuration as a dictionary
    trk_cfg = TrkConfig.load_named(str(config_path.joinpath("tracking_cfg.yaml")))

    # Load the trajectory
    traj = TrajectoryLoader(traj_file)
    print(f"Loaded {traj}")

    # Set up the export -- We'll use the same config set up for both measurements and output of OD process
    cfg = ExportCfg(timestamp=True, metadata={"test key": "test value"})

    # Build the simulated tracking arc, setting the seed to zero
    arc_sim = GroundTrackingArcSim(devices, traj, trk_cfg, 0)
    # Generate the measurements
    msr_path = arc_sim.generate_measurements(
        str(outpath.joinpath("./msr.parquet")), cfg
    )
    print(f"Saved {arc_sim} to {msr_path}")

    # Now let's filter this same data.
    # Load the tracking arc
    arc = DynamicTrackingArc(msr_path)

    # Create the orbit estimate with the covariance diagonal (100 km on position and 1 km/s on velocity)
    orbit_est = OrbitEstimate(
        sc.orbit, covar=np.diag([100.0, 100.0, 100.0, 1.0, 1.0, 1.0])
    )

    msr_noise = [1e-3, 0, 0, 1e-6]
    # Switch from sequential to EKF after 100 measurements
    ekf_num_msr_trig = 100
    # Unless there is a 2 hour gap in the measurements, and then switch back to classical
    ekf_disable_time = Unit.Hour * 2

    rslt_path = process_tracking_arc(
        dynamics["hifi"],
        sc,
        orbit_est,
        msr_noise,
        arc,
        str(outpath.joinpath("./od_result.parquet")),
        cfg,
        ekf_num_msr_trig,
        ekf_disable_time,
        # predict_for=Unit.Hour * 12, # You can predict from the final estimate by uncommenting this line.
    )

    print(f"Stored to {rslt_path}")

    # Load the results
    oddf = pd.read_parquet(rslt_path)
    # Load the reference trajectory
    ref_traj = pd.read_parquet(traj_file)
    # Load the measurements
    msr_df = pd.read_parquet(msr_path)

    # There seems to be a bug when exporting the HTML on Github action windows
    # cf. https://github.com/nyx-space/nyx/actions/runs/5064848025/jobs/9092830624
    if sys.platform != "win32":
        # We'll plot the difference between the reference trajectory and the OD results, with the measurement bands overlaid.
        plot_estimates(
            oddf,
            "OD results from Python",
            cov_frame="RIC",
            ref_traj=ref_traj,
            msr_df=msr_df,
            html_out=str(outpath.joinpath("./od_estimate_plots.html")),
            show=False,
        )

        # Let's also plot the reference and the OD result's orbital elements
        plot_orbit_elements(
            [ref_traj, oddf],
            "Python OD result",
            ["Reference", "OD"],
            html_out=str(outpath.joinpath("./od_vs_ref_elements.html")),
            show=False,
        )

        # More often, the covariance is a better indicator
        plot_covar(
            oddf,
            "OD 1-sigma covar from Python",
            cov_sigma=1.0,
            msr_df=msr_df,
            html_out=str(outpath.joinpath("./od_covar_plots.html")),
            show=False,
        )

        # Now, we'll plot the prefit residuals
        plot_residuals(
            oddf,
            "OD residuals",
            msr_df=msr_df,
            html_out=str(outpath.joinpath("./od_residual_plots.html")),
            show=False,
        )
        # And the postfit histograms
        plot_residual_histogram(
            oddf,
            "OD residuals",
            kind="Postfit",
            html_out=str(outpath.joinpath("./od_residual_histograms.html")),
            show=False,
        )


def test_one_way_msr():
    """
    Test that we can manually perform a one-way measurement
    """

    # Base path
    root = Path(__file__).joinpath("../../../").resolve()
    config_path = root.joinpath("./data/tests/config/")

    # Load the dynamics and spacecraft
    sc = Spacecraft.load(str(config_path.joinpath("spacecraft.yaml")))
    dynamics = SpacecraftDynamics.load_named(str(config_path.joinpath("dynamics.yaml")))

    # Load the devices
    devices = GroundStation.load_many(
        str(config_path.joinpath("./many_ground_stations.yaml"))
    )
    print(f"Loaded {devices}")

    # One way measurement

    end_sc, _ = propagate(sc, dynamics["hifi"], sc.orbit.period() * 1.1)
    print(end_sc)

    range_km, doppler_km_s = devices[0].measure(end_sc.orbit)

    print(range_km, doppler_km_s)
    assert abs(range_km - 18097.562811514355) < 0.1
    assert abs(doppler_km_s - -0.2498238312640348) < 0.1


def test_pure_prediction():
    # Initialize logging
    FORMAT = "%(levelname)s %(name)s %(asctime)-15s %(filename)s:%(lineno)d %(message)s"
    logging.basicConfig(format=FORMAT)
    logging.getLogger().setLevel(logging.INFO)

    # Base path
    root = Path(__file__).joinpath("../../../").resolve()
    config_path = root.joinpath("./data/tests/config/")
    outpath = root.joinpath("output_data/")

    # Load the dynamics and spacecraft
    sc = Spacecraft.load(str(config_path.joinpath("spacecraft.yaml")))
    dynamics = SpacecraftDynamics.load_named(str(config_path.joinpath("dynamics.yaml")))

    # Set up the export -- We'll use the same config set up for both measurements and output of OD process
    cfg = ExportCfg(timestamp=True, metadata={"test key": "test value"})

    # We'll assume that we have a good estimate of the spacecraft's orbit before we predict it forward in time
    # Hence, create the orbit estimate with the covariance diagonal (100 m on position and 50 m/s on velocity)
    orbit_est = OrbitEstimate(
        sc.orbit,
        covar=np.diag([100.0e-3, 100.0e-3, 100.0e-3, 50.0e-3, 50.0e-3, 50.0e-3]),
    )

    rslt_path = predictor(
        dynamics["hifi"],
        sc,
        orbit_est,
        Unit.Second * 15.0,
        str(outpath.joinpath("./od_pred_result.parquet")),
        cfg,
        predict_for=Unit.Hour * 12,
    )

    print(f"Stored to {rslt_path}")

    # Load the prediction results
    oddf = pd.read_parquet(rslt_path)

    # There seems to be a bug when exporting the HTML on Github action windows
    # cf. https://github.com/nyx-space/nyx/actions/runs/5064848025/jobs/9092830624
    if sys.platform != "win32":
        # Let's plot the OD result's orbital elements
        plot_orbit_elements(
            oddf,
            "OD prediction result",
            html_out=str(outpath.joinpath("./od_pred_elements.html")),
            show=False,
        )

        # More often, the covariance is a better indicator
        plot_covar(
            oddf,
            "OD 1-sigma covar from Python",
            cov_sigma=1.0,
            cov_frame="Earth J2000",
            html_out=str(outpath.joinpath("./od_pred_covar_plots.html")),
            show=False,
        )


if __name__ == "__main__":
    test_pure_prediction()



from nyx_space.orbit_determination import GaussMarkov
from nyx_space.time import Unit
from nyx_space.plots import plot_gauss_markov
from pathlib import Path
import pandas as pd


def test_fogm(plot=False):
    """
    Tests a first order Gauss Markov process
    """

    # Base path
    root = Path(__file__).joinpath("../../../").resolve()
    outpath = root.joinpath("output_data/")

    gm = GaussMarkov(tau=Unit.Hour * 24, sigma=5.6, steady_state=0.5)
    print(gm)
    assert (
        str(gm)
        == "First order Gauss-Markov process with τ = 1 days, σ_b = 5.6, σ_q = 0.5"
    )
    gm.simulate(str(outpath.joinpath("fogm.parquet")))

    # Read in the file
    df = pd.read_parquet(str(outpath.joinpath("fogm.parquet")))

    # The seeds are chosen from entropy each time,
    # so the bias should be non-zero but we can't predict what it will be.
    assert df["Bias (unitless)"].mean() != 0.0
    assert df["Bias (unitless)"].max() != 0.0
    assert df["Bias (unitless)"].min() != 0.0
    assert df["Bias (unitless)"].count() == 12525

    if plot:
        plot_gauss_markov(df, title=f"{gm}", tau=gm.tau.to_seconds())


def test_defaults(kinds=["Range", "RangeHP", "Doppler", "DopplerHP"], plot=False):
    """
    Tests the four default models
    """

    # Base path
    root = Path(__file__).joinpath("../../../").resolve()
    outpath = root.joinpath("output_data/")

    for kind in kinds:
        unit = "km" if "Range" in kind else "km/s"
        gm = GaussMarkov.default(kind)
        gm.simulate(str(outpath.joinpath(f"{kind}.parquet")), unit=unit)

        # Read in the file
        df = pd.read_parquet(str(outpath.joinpath(f"{kind}.parquet")))

        if plot:
            plot_gauss_markov(df, title=f"{kind} - {gm}", tau=gm.tau.to_seconds())


def test_load(plot=False):
    """
    Tests that we can load a Gauss Markov process from a file and simulate it.
    """
    # Base path
    root = Path(__file__).joinpath("../../../").resolve()
    config = root.joinpath("data/tests/config/high-prec-network.yaml")
    outpath = root.joinpath("output_data/")

    models = GaussMarkov.load_named(str(config))

    for name, gm in models.items():
        gm.simulate(str(outpath.joinpath(f"{name}.parquet")))

        # Read in the file
        df = pd.read_parquet(str(outpath.joinpath(f"{name}.parquet")))

        if plot:
            plot_gauss_markov(df, title=f"{name} - {gm}", tau=gm.tau.to_seconds())


if __name__ == "__main__":
    test_fogm(True)
    test_defaults(plot=True)
    test_load(True)



"""
Nyx, blazing fast astrodynamics
Copyright (C) 2023 Christopher Rabotin <christopher.rabotin@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import plotly.express as px

from .utils import finalize_plot


def plot_gauss_markov(df, title="Gauss Markov Process", tau=None):
    """
    Plots a Gauss Markov process from the provided data frame
    """

    # Grab the column
    try:
        col_name = [col for col in df.columns if "Bias" in col][0]
    except IndexError:
        raise ValueError("No bias column found in the provided data frame")

    fig = px.scatter(
        df,
        x="Delta Time (s)",
        y=col_name,
        color="Run",
        opacity=0.4,
        marginal_y="rug",
    )

    # TODO: Consider adding the autocorrelation data for clarity

    if tau:
        fig.add_vline(
            x=tau,
            line_width=2,
            line_dash="dash",
            line_color="red",
            row=1,
            col=1,
        )

    finalize_plot(fig, title=title)

    fig.show()



"""
Nyx, blazing fast astrodynamics
Copyright (C) 2023 Christopher Rabotin <christopher.rabotin@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import plotly.graph_objects as go
from plotly.subplots import make_subplots
import pandas as pd

from .utils import (
    radii,
    build_sphere,
    build_3dline,
    colors,
    finalize_plot,
    body_color,
    nyx_tpl,
)


def plot_traj(
    dfs,
    title,
    html_out=None,
    copyright=None,
    fig=None,
    center="Earth",
    show=True,
):
    """
    Plot a trajectory in 3D

    Args:
        dfs (pandas.DataFrame): The data frame containing the trajectory (or a list thereof)
        title (str): The title of the plot
        html_out (str, optional): The path to save the HTML to. Defaults to None.
        copyright (str, optional): The copyright to display on the plot. Defaults to None.
        fig (plotly.graph_objects.Figure, optional): The figure to add the trajectory to. Defaults to None.
        center (str, optional): The name of the center object, e.g. `Luna` (Moon). Defaults to "Earth".
        show (bool, optional): Whether to show the plot. Defaults to True. If set to false, the figure will be returned.
    """

    if fig is None:
        fig = go.Figure()

    # Only plot the sphere if there currently is nothing on the plot
    if len(fig.data) == 0:
        try:
            traces = [build_sphere(radii[center], body_color[center], opacity=0.8)]
        except KeyError:
            print("Body must be one of:" + ", ".join(radii.keys()))
            raise
    else:
        traces = []

    if not isinstance(dfs, list):
        dfs = [dfs]

    for df in dfs:
        print(df.describe())
        print(df.columns)
        color_values = list(colors.values())
        traces += [
            build_3dline(
                df["x (km)"],
                df["y (km)"],
                df["z (km)"],
                df["Epoch:Gregorian UTC"],
                color=color_values[len(fig.data) % len(color_values)],
                name=title,
            )
        ]

    # Now that we have the data, let's plot it
    fig.add_traces(traces)
    finalize_plot(fig, title, copyright=copyright)
    fig.update_layout(scene_aspectmode="data")

    if html_out:
        with open(html_out, "w") as f:
            f.write(fig.to_html())
        print(f"Saved HTML to {html_out}")

    if show:
        fig.show()
    else:
        return fig


def plot_ground_track(
    dfs,
    title,
    names,
    projection="orthographic",
    landmarks={},
    not_earth=False,
    html_out=None,
    copyright=None,
    fig=None,
    show=True,
):
    """
    Plot a ground track

    Args:
        dfs (pandas.DataFrame): The data frame containing the trajectory (or a list thereof)
        title (str): The title of the plot
        names (str): The name of the trajectory (or a list thereof)
        projection (str, optional): The projection to use. Defaults to "orthographic". Another good option is "equirectangular"
        landmarks (dict, optional): A dictionary of landmarks to plot. Defaults to {}. Example: {"Eiffel Tower": (48.8584, 2.2945)} (latitude, longitude)
        not_earth (bool, optional): Whether to plot the Earth. Defaults to False.
        html_out (str, optional): The path to save the HTML to. Defaults to None.
        copyright (str, optional): The copyright to display on the plot. Defaults to None.
        fig (plotly.graph_objects.Figure, optional): The figure to add the trajectory to. Defaults to None.
        show (bool, optional): Whether to show the plot. Defaults to True. If set to false, the figure will be returned.
    """
    if fig is None:
        fig = go.Figure()

    color_values = list(colors.values())

    if not isinstance(dfs, list):
        dfs = [dfs]

    if not isinstance(names, list):
        names = [names]

    assert len(names) == len(dfs), "Number of names must match number of dataframes"

    for name, df in zip(names, dfs):
        i = len(fig.data)
        try:
            color = color_values[i % len(color_values)]
            fig.add_trace(
                go.Scattergeo(
                    lon=df["geodetic_longitude (deg)"],
                    lat=df["geodetic_latitude (deg)"],
                    mode="lines",
                    name=name,
                    text=df["Epoch:Gregorian UTC"],
                    line=dict(
                        color=f"rgb({int(color[0])}, {int(color[1])}, {int(color[2])})",
                    ),
                )
            )
        except KeyError as e:
            print(
                "\n\nLatitude and longitude not found in the data.\n"
                "Generate the output with traj.to_csv_with_groundtrack(...), \n"
                'Or specify `"geodetic_latitude", "geodetic_longitude"` in the output headers\n\n'
            )
            raise e

    for name, (lat, long) in landmarks.items():
        print(f"Adding landmark {name} at lat={lat}, long={long}")
        fig.add_trace(
            go.Scattergeo(lon=[float(long)], lat=[float(lat)], name=name.strip())
        )

    fig.update_geos(
        lataxis_showgrid=True,
        lonaxis_showgrid=True,
        projection_type=projection,
        visible=not not_earth,
    )
    fig.update_layout(margin={"r": 0, "t": 0, "l": 0, "b": 0})

    finalize_plot(fig, f"Ground track: {title}", copyright=copyright)

    if html_out:
        with open(html_out, "w") as f:
            f.write(fig.to_html())
        print(f"Saved HTML to {html_out}")

    if show:
        fig.show()
    else:
        return fig


def plot_orbit_elements(
    dfs,
    title,
    names=[],
    html_out=None,
    copyright=None,
    show=True,
):
    """
    Plots the orbital elements: SMA, ECC, INC, RAAN, AOP, TA, True Longitude, AOL

    Args:
        dfs (pandas.DataFrame): The data frame containing the trajectory (or a list thereof)
        title (str): The title of the plot
        html_out (str, optional): The path to save the HTML to. Defaults to None.
        copyright (str, optional): The copyright to display on the plot. Defaults to None.
        fig (plotly.graph_objects.Figure, optional): The figure to add the trajectory to. Defaults to None.
        center (str, optional): The name of the center object, e.g. `Luna` (Moon). Defaults to "Earth".
        show (bool, optional): Whether to show the plot. Defaults to True. If set to false, the figure will be returned.
    """

    if not isinstance(dfs, list):
        dfs = [dfs]

    for df in dfs:
        pd_ok_epochs = []
        for epoch in df["Epoch:Gregorian UTC"]:
            epoch = epoch.replace("UTC", "").strip()
            if "." not in epoch:
                epoch += ".0"
            pd_ok_epochs += [epoch]
        df["Epoch"] = pd.to_datetime(pd_ok_epochs)

    if not isinstance(names, list):
        names = [names]

    if len(names) > 1:
        assert len(names) == len(dfs), "Number of names must match number of dataframes"

    columns = [
        "sma (km)",
        "ecc",
        "inc (deg)",
        "raan (deg)",
        "aop (deg)",
        "ta (deg)",
        "aol (deg)",
        "tlong (deg)",
    ]

    fig = make_subplots(rows=4, cols=2, subplot_titles=columns)

    row_i = 0
    col_i = 0

    for col in columns:
        for k, df in enumerate(dfs):
            try:
                name = f"{names[k]} {col}"
            except IndexError:
                name = col

            fig.add_trace(
                go.Scatter(x=df["Epoch"], y=df[col], name=name),
                row=row_i + 1,
                col=col_i + 1,
            )
        col_i = (col_i + 1) % 2
        if col_i == 0:
            row_i = (row_i + 1) % 4

    finalize_plot(fig, title, copyright=copyright)

    if html_out:
        with open(html_out, "w") as f:
            f.write(fig.to_html())
        print(f"Saved HTML to {html_out}")

    if show:
        fig.show()
    else:
        return fig



"""
Nyx, blazing fast astrodynamics
Copyright (C) 2023 Christopher Rabotin <christopher.rabotin@gmail.com>

This program is free software: you can redistribute it and/or modify
it under the terms of the GNU Affero General Public License as published
by the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU Affero General Public License for more details.

You should have received a copy of the GNU Affero General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

from . import _nyx_space, plots, utils

from nyx_space import time, cosmic, mission_design, orbit_determination

__all__ = ["time", "cosmic", "mission_design", "orbit_determination", "plots"]



import logging
from pathlib import Path
import pickle

from nyx_space.cosmic import Spacecraft, Orbit, SrpConfig, Cosm
from nyx_space.mission_design import (
    propagate,
    StateParameter,
    Event,
    SpacecraftDynamics,
    TrajectoryLoader,
)
from nyx_space.time import Duration, Unit, Epoch


def test_propagate():
    # Initialize logging
    FORMAT = "%(levelname)s %(name)s %(filename)s:%(lineno)d\t%(message)s"
    logging.basicConfig(format=FORMAT)
    logging.getLogger().setLevel(logging.INFO)

    # Base path
    root = Path(__file__).joinpath("../../../").resolve()

    config_path = root.joinpath("./data/tests/config/")

    sc = Spacecraft.load(str(config_path.joinpath("spacecraft.yaml")))
    # Check that we have loaded this correctly by checking the values from the YAML file
    assert sc.value_of(StateParameter.X) == -9042.862234
    assert sc.value_of(StateParameter.Y) == 18536.333069
    assert sc.value_of(StateParameter.Z) == 6999.957069
    assert sc.value_of(StateParameter.VX) == -3.288789
    assert sc.value_of(StateParameter.VY) == -2.226285
    assert sc.value_of(StateParameter.VZ) == 1.646738
    assert sc.value_of(StateParameter.Cr) == 1.0
    assert sc.value_of(StateParameter.Cd) == 2.2
    assert sc.value_of(StateParameter.FuelMass) == 50.0
    assert sc.epoch.timedelta(Epoch("2018-09-15T00:15:53.098 UTC")) == Duration.zero()
    # Check other stuff that is computed on request
    assert sc.value_of(StateParameter.SMA) == 21999.99774705774

    dynamics = SpacecraftDynamics.load_named(str(config_path.joinpath("dynamics.yaml")))
    # So far, there are no accessors on the dynamics, so we will check what they print out =(
    assert (
        f"{dynamics['lofi']}"
        == "Spacecraft dynamics (with guidance = false): No force models; Orbital dynamics: Point masses of Sun J2000, Earth J2000"
    )
    assert (
        f"{dynamics['hifi']}"
        == "Spacecraft dynamics (with guidance = false): SRP with φ = 1367 W/m^2 and eclipse light-source: Sun J2000, shadows casted by: Sun J2000, Moon J2000;  Orbital dynamics: Point masses of Sun J2000, Earth J2000, Moon J2000; IAU Earth gravity field 10x10 (order x degree)"
    )

    rslt, traj = propagate(sc, dynamics["lofi"], Unit.Day * 5.159)
    # Check that we propagated for the correct duration
    assert rslt.epoch.timedelta(sc.epoch) == Duration("5 days 3 h 48 min 57 s 600 ms")
    # We aren't in two body dynamics so the SMA has changed among other things.
    assert rslt.value_of(StateParameter.SMA) == 22000.024381883788
    # Grab a specific epoch from the trajectory
    sc_state = traj.at(Epoch("2018-09-16T00:16:53 TDB"))
    # Check that we got the right epoch
    assert sc_state.epoch.timedelta(Epoch("2018-09-16T00:16:53 TDB")) == Duration.zero()
    # Save the file to parquet
    traj.to_parquet("lofi.parquet")

    # We can also propagate with a different method
    rslt, traj = propagate(sc, dynamics["lofi"], Unit.Day * 5.159, method="Dormand78")
    assert rslt.epoch.timedelta(sc.epoch) == Duration("5 days 3 h 48 min 57 s 600 ms")

    # Let's now propagate the original spacecraft to its apoapsis, but let's search no more than 2 orbit periods
    rslt_apo, _ = propagate(
        sc,
        dynamics["lofi"],
        sc.orbit.period() * 2,
        event=Event(StateParameter.Apoapsis, 0.0),
    )
    # Let's make sure that worked to within the default precision
    assert abs(rslt_apo.value_of(StateParameter.TrueAnomaly) - 180.0) < 0.001

    event = Event(StateParameter.Apoapsis, 0.0, value_precision=1e-6)

    # But we can also be even more precise
    rslt_apo, traj = propagate(
        sc,
        dynamics["lofi"],
        sc.orbit.period() * 2,
        event=event,
    )
    assert abs(rslt_apo.orbit.ta_deg() - 180.0) <= 1e-6

    # Resample the trajectory at fixed step size of 25 seconds
    traj = traj.resample(Unit.Second * 25.0)

    # Export this trajectory with additional metadata and the events
    traj.to_parquet(
        "lofi_with_events.parquet", metadata={"test key": "test value"}, events=[event]
    )

    # Also export this ground track
    cosm = Cosm.de438()
    iau_earth = cosm.frame("IAU Earth")
    traj.to_parquet("iau_earth_lofi.parquet", groundtrack=iau_earth)

    # Let's also search for this event in the trajectory
    for sc_at_event in traj.find(event):
        print(sc_at_event)
        assert abs(sc_at_event.value_of(StateParameter.TrueAnomaly) - 180.0) <= 1e-6


def test_build_spacecraft():
    """
    Tests that we can build a spacecraft from scratch without an input file
    """
    # Initialize logging
    FORMAT = "%(levelname)s %(name)s %(filename)s:%(lineno)d\t%(message)s"
    logging.basicConfig(format=FORMAT)
    logging.getLogger().setLevel(logging.INFO)

    cosm = Cosm.de438()
    eme2k = cosm.frame("EME2000")  # Earth Mean Equator J2000
    orbit = Orbit.from_keplerian_altitude(
        400.0, 0.01, 15.6, 45.0, 90.0, 75.0, Epoch.system_now(), eme2k
    )
    # Define the SRP
    srp = SrpConfig(2.0)  # 2.0 will be the area
    print(srp)
    assert srp.cr == 1.8  # Default value
    assert srp.area_m2 == 2.0
    sc = Spacecraft(orbit, 150.0, 15.0, srp)
    print(sc)

    assert sc.srp().__eq__(srp)  # Not sure why the `==` operator doesn't work here
    assert sc.drag().area_m2 == 0.0
    assert (
        sc.drag().cd == 2.2
    )  # Default value, but the area is zero, so it doesn't have any effect

    # Using this spacecraft as a template, let's load an OEM file, convert it to Parquet, and ensure we can load it back in.
    # The orbit data will be overwritten with data from the OEM file.
    root = Path(__file__).joinpath("../../../").resolve()

    config_path = root.joinpath("./data/tests/ccsds/oem/LEO_10s.oem")
    output_path = root.joinpath("./output_data/LEO_10s.parquet")
    # Convert from OEM to Parquet will happen on load
    traj = TrajectoryLoader(str(config_path), "oem", str(output_path), sc)
    print(traj)
    # Check that we can pickle the trajectory loader object
    traj_pkl = pickle.dumps(traj)
    traj_unpkl = pickle.loads(traj_pkl)
    assert traj_unpkl.__eq__(traj)


if __name__ == "__main__":
    # test_propagate()
    test_build_spacecraft()
